#include "database.h"
#include <fstream>
#include <iostream>
#include <algorithm>
#include <functional>
#include <cwctype>
#include <vector>
#include <sstream>
#include <iterator>
#include <locale>
#include <cwchar>

using namespace std;

bool RBST::add(const Item &key, bool verbose)
{
	unsigned int oldSize = m_size;
	node *newNode = new node(key);
	m_root = randomAdd(m_root, newNode);
	if (m_size > oldSize)
	{
		if (verbose)
			wcout << "Node " << key << " is added into the tree." << endl;

		return true;
	}
	else
	{
		if (verbose)
			wcout << "Node " << key << " is already in the tree." << endl;

		return false;
	}
};

bool RBST::del(const Item &key, bool verbose)
{
	unsigned oldSize = m_size;
	m_root = del(m_root, key);
	if (m_size < oldSize)
	{
		if (verbose)
			wcout << "Node " << key << " is deleted from the tree." << endl;

		return true;
	}
	else
	{
		if (verbose)
			wcout << "Node " << key << " is not in the tree." << endl;

		return false;
	}
};

bool RBST::find(const Item &key, bool verbose) const
{
	node* ret = find(m_root, key);
	if (ret == NULL)
	{
		if (verbose)
			wcout << "Node " << key << " is not in the tree." << endl;

		return false;
	}
	else
	{
		if (verbose)
			wcout << "Node " << key << " is in the tree." << endl;

		return true;
	}
};

node* RBST::find(node* target, std::function<int(const Item &k)> f) const
{
	if (target == NULL)
		return NULL;

	int comparison = f(target->getKey());
	
	if (comparison < 0)
		return find(target->left(), f);
	else if (comparison == 0)
		return target;
	else
		return find(target->right(), f);

	return NULL;
}

node* RBST::findFromRoot(std::function<int(const Item &k)> f) const
{
	return find(m_root, f);
}

node*  RBST::rightRotate(node* target)
{
	node *l = target->left();
	target->setLeft(l->right());
	l->setRight(target);
	return l;
};

node*  RBST::leftRotate(node* target)
{
	node *r = target->right();
	target->setRight(r->left());
	r->setLeft(target);
	return r;
};

node* RBST::addRoot(node* target, node* newNode)
{
	if (target == NULL)
		return newNode;

	int comparison = newNode->compare(*target);

	if (comparison < 0)
	{
		target->setLeft(addRoot(target->left(), newNode));
		return rightRotate(target);
	}
	else
	{
		target->setRight(addRoot(target->right(), newNode));
		return leftRotate(target);
	}
};

node* RBST::randomAdd(node* target, node* newNode)
{
	if (target == NULL)
	{
		m_size++;
		newNode->setSize(1);
		return newNode;
	}

	int r = rand() % target->getSize() + 1;

	int comparison = newNode->compare(*target);

	if (comparison == 0) return target;

	if (r == 1)
	{
		newNode->setSize(target->getSize() + 1);
		m_size++;
		return addRoot(target, newNode);
	}

	if (comparison < 0)
		target->setLeft(randomAdd(target->left(), newNode));
	else if (comparison > 0)
		target->setRight(randomAdd(target->right(), newNode));

	target->setSize(target->getSize() + 1);

	return target;
};

node* RBST::del(node* target, Item key)
{
	if (target == NULL)
		return NULL;

	int comparison;
	if (key < target->getKey())
		comparison = -1;
	else if (key == target->getKey())
		comparison = 0;
	else
		comparison = 1;

	if (comparison == 0) { 			// Equal case
		target = deleteNode(target);
		m_size--;
	}
	else if (comparison < 0) 		// key < target
		target->setLeft(del(target->left(), key));
	else							// key > target
		target->setRight(del(target->right(), key));

	return target;
};

node* RBST::deleteNode(node* target)
{
	if (target->right() == NULL)
		return target->left(); 		// 0 or 1 child case
	else if (target->left() == NULL)
		return target->right();		// 1 child case

	else
	{
		node *min = findMin(target->right());
		min->setRight(removeMinLeaf(target->right(), min));
		min->setLeft(target->left());
		return min;
	}
};

node* RBST::removeMinLeaf(node* target, node* min)
{
	if (target->compare(*min) == 0)
		return min->right();
	else
		target->setLeft(removeMinLeaf(target->left(), min));
	return target;
};

node* RBST::findMin(node* t) const
{
	node* ret = t;
	while (ret->left() != NULL)
		ret = ret->left();

	return ret;
}

node* RBST::find(node* target, Item key) const
{
	if (target == NULL)
		return NULL;

	int comparison;
	if (key < target->getKey())
		comparison = -1;
	else if (key == target->getKey())
		comparison = 0;
	else
		comparison = 1;

	if (comparison < 0)
		return find(target->left(), key);
	else if (comparison == 0)
		return target;
	else
		return find(target->right(), key);

	return NULL;
}

Item DataBase::find(const std::wstring &k) const
{
	auto compfunc = [&k](const Item &item) {
		if (item > k)
			return -1;
		else if (item == k)
			return 0;
		else
			return 1;
	};
	node* res = tree.findFromRoot(compfunc);
	if (res != NULL)
		return res->getKey();

	return Item();
}

void DataBase::findAllLike(const std::wstring &k, std::vector<Item> &result) const
{
	auto compfunc = [&k](const Item &i) {
		wstring s(i.c_str(), 3);
		for (std::wstring::iterator it = s.begin(); it != s.end(); it++)
			(*it) = tolower((*it), locale(loc));

		if (s > k)
			return -1;
		else if (s == k)
			return 0;
		else
			return 1;
	};

	auto addToRes = [&result](node* n) {
		if (n != NULL)
			result.push_back(n->getKey());
	};
	node* first = tree.findFromRoot(compfunc);
	if (first == NULL)
		return;

	addToRes(first);
	
	node* n = first;
	do {
		n = tree.find(n->left(), compfunc);
		addToRes(n);
	} while (n != NULL);

	n = first;
	do {
		n = tree.find(n->right(), compfunc);
		addToRes(n);
	} while (n != NULL);
}

#define ITEMLEN 64

bool loadOnlyFname(wistream &in, wstring &item)
{
	wchar_t buff[ITEMLEN + 1];
	const wchar_t *cur = buff;
	in.read(buff, ITEMLEN);
	if (in.gcount() < ITEMLEN)
		return false;
	
	wchar_t* f = wcstok(buff, L" ");
	item = wstring(f);
	return true;
}

void DataBase::loadFromFile(const std::string &fileName)
{
	locale l(loc);
	wcout.imbue(l);
	wifstream in(fileName, ios::in | ios::binary);
	in.imbue(l);

	Item item;
	//while (in >> item)
	while (loadOnlyFname(in, item))
		tree.add(item);

	cout << tree.size() << " items loaded from file: " << fileName << endl;
}